//
// Copyright (C) 2005 Andras Varga
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
//

#include <vector>
#include <omnetpp.h>
#include <string.h>
#include "Valores.h"

#include "Tdelay.h"


#include <iostream>
#include <fstream>
using namespace std;

Define_Module(Tdelay);


void Tdelay::initialize()
{
    startTime = par("startTime");
    long _batchSize = par("batchSize");
    if((_batchSize < 0) || (((long)(unsigned int)_batchSize) != _batchSize))
        throw cRuntimeError("invalid 'batchSize=%ld' parameter at '%s' module",_batchSize, getFullPath().c_str());
    batchSize = (unsigned int)_batchSize;
    maxInterval = par("maxInterval");

    emisor = par("emisor");
    numPackets = numBits = 0;
    intvlStartTime = intvlLastPkTime = 0;
    intvlNumPackets = intvlNumBits = 0;
    seqnum=0;
    expectedReplySeqNo=0;
    outoftimeReplySeqNo=0;
    tempexpired=0;
    outOfOrderArrivalCount=0;
    timeoutEvent=NULL;

    WATCH(numPackets);
    WATCH(numBits);
    WATCH(intvlStartTime);
    WATCH(intvlNumPackets);
    WATCH(intvlNumBits);

    bitpersecVector.setName("thruput (bit/sec)");
    pkpersecVector.setName("packet/sec");
    pkDelayVector.setName("pkDelayVector");
    pkDelayStats.setName("pkDelayStats");
    JitterVector.setName("JitterVector");

}

void Tdelay::handleMessage(cMessage *msg)
{
	//if its a timer
	 if (msg->isSelfMessage())
	 {
		 tempexpired=1;


	 }
	 else
	 {
			if (msg->hasObject(""))
			{
				 Valores *nuevo =(Valores *) msg->getObject("");

				 ev << "  ee1:    " << nuevo->getee1() << endl;
				 ev << "  simtime:    " << nuevo->getsimtime1() << endl;
				 ev << "  Snum    " << nuevo->getsnum() << endl;


				 updateStats(simTime(), PK(msg)->getBitLength());


				 if (nuevo->getsnum() == expectedReplySeqNo)
				 {
					 // expected ping reply arrived; expect next sequence number
					 expectedReplySeqNo++;

					 if (timeoutEvent!=NULL)
					{
					   cancelEvent(timeoutEvent);
					   delete(timeoutEvent);
					   timeoutEvent = NULL;

					}

				   	timeoutEvent = new cMessage("timeout");
					 scheduleAt(simTime()+2.0, timeoutEvent);
					 tempexpired=0;

					 RecordDelay(nuevo);
				 }
				 else if (nuevo->getsnum() > expectedReplySeqNo)
				 {
					 EV << "Jump in seq numbers, assuming pings since #" << expectedReplySeqNo << " got lost\n";
					 outoftimeReplySeqNo=expectedReplySeqNo;
					 // expect sequence numbers to continue from here
					 expectedReplySeqNo = nuevo->getsnum()+1;


					if (timeoutEvent!=NULL)
					{
						cancelEvent(timeoutEvent);
						delete(timeoutEvent);
						timeoutEvent = NULL;

					}
					timeoutEvent = new cMessage("timeout");
					scheduleAt(simTime()+2.0, timeoutEvent);
					tempexpired=0;
					 RecordDelay(nuevo);
				 }
				 else // seqNo < expectedReplySeqNo
				 {
					 if(tempexpired==1 || nuevo->getsnum() < outoftimeReplySeqNo)
					 {
						 // ping arrived too late: count as out of order arrival
						 EV << "Arrived out of order (too late)\n";
						 outOfOrderArrivalCount++;

					 }
					 else
						 RecordDelay(nuevo);
				 }




			}

	 }


    delete(msg);

}

void Tdelay::RecordDelay(Valores *nuevo)
{

    pkJitter.collect(simTime() - nuevo->getsimtime1());


	if(pkJitter.getCount()>1)
    eej=pkJitter.getStddev();
	else
		eej=0;
    JitterVector.record(eej);
    pkJitter2.collect(eej);

    pkDelayVector.record(simTime() - nuevo->getsimtime1()+ nuevo->getee1() );


    pkDelayStats.collect(simTime() - nuevo->getsimtime1()+ nuevo->getee1());
}

void Tdelay::updateStats(simtime_t now, unsigned long bits)
{
    numPackets++;
    numBits += bits;

    // packet should be counted to new interval
    if (intvlNumPackets >= batchSize || now-intvlStartTime >= maxInterval)
        beginNewInterval(now);

    intvlNumPackets++;
    intvlNumBits += bits;
    intvlLastPkTime = now;
}

void Tdelay::beginNewInterval(simtime_t now)
{
    simtime_t duration = now - intvlStartTime;

    // record measurements
    double bitpersec = intvlNumBits/duration.dbl();
    double pkpersec = intvlNumPackets/duration.dbl();

    bitpersecVector.recordWithTimestamp(intvlStartTime, bitpersec);
    pkpersecVector.recordWithTimestamp(intvlStartTime, pkpersec);

    // restart counters
    intvlStartTime = now;  // FIXME this should be *beginning* of tx of this packet, not end!
    intvlNumPackets = intvlNumBits = 0;
}

void Tdelay::finish()
{
    simtime_t duration = simTime() - startTime;

    ev << "  pk. delay, min:    " << pkDelayStats.getMin() << endl;
    ev << "  pk. delay, max:    " << pkDelayStats.getMean() << endl;
    ev << "  pk. delay, mean:   " << pkDelayStats.getMax() << endl;
    ev << "  pk. delay, stddev: " << pkDelayStats.getStddev() << endl;



    recordScalar("JitterMean", pkJitter2.getMean());
    recordScalar("EedMean",pkDelayStats.getMean());
    recordScalar("duration", duration);
    recordScalar("Drop packets", outOfOrderArrivalCount);
    recordScalar("total packets", numPackets);
    recordScalar("total bits", numBits);

    recordScalar("avg throughput (bit/s)", numBits/duration.dbl());
    recordScalar("avg packets/s", numPackets/duration.dbl());
}


